/*
 * "Pimp My Wreck" - Lousy Car Pimpin' kit by Yaron Grushka and Tomet Novick.
 * The following sketch controls our kit which includes:
 * 1. Speed dial - make calls without having to touch your phone using an IR remote, via Blynk server and Integromat scenarios.
 * 2. Interior lighting - an Adafruit NePixel 8x8 LED matrix, also controlled with the IR remote, replaces your car's old-fashioned interior lighting.
 *    A piezo sensor controls the color of the lights.
 * 3. Parking location - A reed switch that closes when you park your car prompts Blynk to send you your parking location via email so you don't lose your car.
 *    You can also attach reminders to the email via a terminal widget in the Blynk dashboard. An Adafruit IO dashboard gives a real-time indication of whether
 *    your car is currently parked, and also displays your last known parking location on a map.
 * 4. Phone wake-up mechanism - a servo motor helps keep your smartphone ready for action by tapping it in regular intervals, so it doesn't lock.
*/
#define BLYNK_PRINT Serial
#define BLYNK_MAX_SENDBYTES 1200 // Allows for sending longer emails, if needed

#ifndef UNIT_TEST
#include <Arduino.h>
#endif
#include <IRremoteESP8266.h>
#include <IRrecv.h>
#include <IRutils.h>
#include <BlynkSimpleEsp8266.h>
#include <Servo.h>
#include <Adafruit_NeoPixel.h>
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"
#include <stdlib.h>

// All the input and output pins that we're using
uint16_t RECV_PIN = D2;     // IR receiver
uint16_t SMALL_SERVO = D3;  // Servo motor
uint16_t PARKING = D4;      // Reed switch
uint16_t LED = D5;          // LED to indicate IR signals
uint16_t LED2 = D6;         // LED to indicate parking
uint16_t MATRIX = D7;       // LED matrix
uint16_t PIEZO = A0;        // Piezo sensor

// WiFi credentials
char ssid[] = "Your network name here";
char pass[] = "Your network password here";

// Blynk authorization code
char auth[] = "Your Blynk auth code here";

// Adafruit IO information and setup
#define AIO_SERVER "io.adafruit.com"
#define AIO_SERVERPORT 1883
#define AIO_USERNAME "Your Adafruit IO username here"
#define AIO_KEY "Your secret Adafruit IO key here"

void connect();
WiFiClient client;
const char MQTT_CLIENTID[] = AIO_KEY __DATE__ __TIME__;
const char MQTT_USERNAME[] = AIO_USERNAME;
const char MQTT_PASSWORD[] = AIO_KEY;
const char MQTT_SERVER[] = AIO_SERVER;
Adafruit_MQTT_Client mqtt(&client, MQTT_SERVER, AIO_SERVERPORT, MQTT_CLIENTID, MQTT_USERNAME, MQTT_PASSWORD);

// Parking feed
const char PARKED_FEED[] = AIO_USERNAME "/feeds/parked";
Adafruit_MQTT_Publish parked = Adafruit_MQTT_Publish(&mqtt, PARKED_FEED);
int carParked = 1;

// Parking location feed
const char LOCATION_FEED[] = AIO_USERNAME "/feeds/location/csv";
Adafruit_MQTT_Publish loc = Adafruit_MQTT_Publish(&mqtt, LOCATION_FEED);

// Everything we need for the IR receiver
IRrecv irrecv(RECV_PIN);
decode_results results;
boolean callMade = false;
long ledInterval = 100;
long previousLedMillis = 0;

// Everything we need for the servo
Servo smallServo;
long interval = 6000; // We want our servo to move every 6 seconds (6000 milliseconds). If another interval suits you better, feel free to change it.
long previousServoMillis = 500;
boolean angle = true;

// Everything we need for the piezo
long previousPiezoMillis = 0;
int piezoVal = 0;

// GPS location variables
float latitude = 0;
float longitude = 0;
int alt = 0;
float s = 0;

// Terminal message variables
WidgetTerminal terminal(V11);
String noteToSelf = "";

// True iff the car is parked
boolean isParked = true;

// Everything we need for the matrix
Adafruit_NeoPixel strip = Adafruit_NeoPixel(64, MATRIX, NEO_GRB + NEO_KHZ800);
uint32_t colors[] = {strip.Color(100, 100, 100), strip.Color(100, 0, 0), strip.Color(0, 100, 0), strip.Color(0, 0, 100),
                     strip.Color(100, 100, 0), strip.Color(100, 0, 100), strip.Color(0, 100, 100)
                    };
uint32_t color = strip.Color(100, 100, 100); // Default color (white)
int randNumber = 0;
int prevNumber = 0;
boolean isLit = false;


// This function takes our note-to-self from the Blynk terminal
BLYNK_WRITE(V11) {
  if (String("") != param.asStr()) {
    noteToSelf += param.asStr();
    noteToSelf += "\n";
  }
}

void setup() {
  Serial.begin(9600);
  Blynk.begin(auth, ssid, pass);      // Activate Blynk
  irrecv.enableIRIn();                // Activate the IR receiver
  smallServo.attach(SMALL_SERVO);     // Activate the servo
  strip.begin();                      // Activate the LED matrix
  pinMode(PARKING, INPUT_PULLUP);     // Initialize the Reed switch
  pinMode(PIEZO, INPUT_PULLUP);       // Initialize the piezo
  pinMode(LED, OUTPUT);               // Initialize the two small LED bulbs
  pinMode(LED2, OUTPUT);

  // Setting connection to Adafruit IO
  WiFi.begin(ssid, pass);
  // wait for WiFi connection
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
    delay(10);
  }
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());
  Serial.println();
  // connect to adafruit io
  connect();

}

void loop() {
  Blynk.run();

  // Goes off when vehicle was parked
  if (digitalRead(PARKING) == LOW) { // Reed switch is closed
    if (!isParked) { // This means we just switched to paking
      isParked = true;

      char buff[10];
      String msg = "";
      if (String("") != noteToSelf) { // If we sent ourselves a note from the Blynk terminal, send it first. Else, send only the parking coordinates.
        msg += "NOTE TO SELF:\n";
        msg += noteToSelf;
        msg += "\n";
      }

      // Parse latitude and longitude to the string that will be emailed to us
      msg += "Your parking coordinates are:\nLat: ";
      dtostrf(latitude, 4, 6, buff); // Parse latitude to a char buffer
      msg += buff;
      msg += "\nLon: ";
      dtostrf(longitude, 4, 6, buff); // Parse longitude to a char buffer
      msg += buff;
      msg += "\nInsert lat and lon coordinates at: https://www.google.com/maps.";

      Blynk.email("Your parking coordinates", msg); // Send ourselves the email

      // Send parking data to Adafruit IO
      carParked = 1;
      if (!parked.publish(carParked)) Serial.println("Failed");
      else Serial.println("Sent OK! " + String(carParked));

      // Parse GPS location data into CSV format
      char GPSBuffer[30]; // Combine Latitude, Longitude, Altitude into a buffer of size X
      char *p = GPSBuffer; // Create a buffer to store GPS information to upload to Adafruit IO

      dtostrf(s, 1, 4, p); // Convert speed to a String Variable and add it to the buffer
      p += strlen(p);
      p[0] = ','; p++;
      
      dtostrf(latitude, 3, 6, p); // Convert latitude to a String variable and add it to the buffer
      p += strlen(p);
      p[0] = ','; p++;
      
      dtostrf(longitude, 3, 6, p); // Convert longitude to a String variable and add it to the buffer
      p += strlen(p);
      p[0] = ','; p++;
      int altAsInt = alt;
      
      dtostrf(alt, 1, 0, p); // Convert altitude to a String variable and add it to the buffer
      p += strlen(p);
      
      p[0] = 0; // null terminate, end of buffer array

      // Send GPS location data to Adafruit IO
      if (!loc.publish(GPSBuffer)) Serial.println("Failed");
      else{
        Serial.print("Sent OK! ");
        Serial.println(GPSBuffer);
      }

      digitalWrite(LED2, HIGH);   // Small LED that indicates that we've parked our car
    }
  } else {
    digitalWrite(LED2, LOW);
    if (isParked) {
      isParked = false;
      // Send data to Adafruit IO
      carParked = 0;
      if (!parked.publish(carParked)) Serial.println("Failed");
      else Serial.println("Sent OK! " + String(carParked));
    }
  }

  // Here we activate our servo that keeps the cellphone from locking. The servo taps the smartphone every 6 seconds.
  unsigned long currentMillis = millis();
  if (currentMillis - previousServoMillis >= interval) {
    if (angle) {
      previousServoMillis = currentMillis;
      smallServo.write(120);
      interval = 1000;
      angle = false;
    } else {
      previousServoMillis = currentMillis;
      smallServo.write(90);
      interval = 6000;
      angle = true;
    }
  }

  // Light the small LED for 100 millis when the IR sensor receives input
  if (callMade && (currentMillis - previousLedMillis >= ledInterval)) {
    digitalWrite(LED, LOW);
    callMade = false;
  }

  if (irrecv.decode(&results)) {
    unsigned int value = results.value;
    switch (value) {
      case 16753245: // "1" button, speed dial 1
        Blynk.virtualWrite(V1, 1);
        digitalWrite(LED, HIGH);
        callMade = true;
        previousLedMillis = currentMillis;
        break;

      case 16736925: // "2" button, speed dial 2
        Blynk.virtualWrite(V2, 1);
        digitalWrite(LED, HIGH);
        callMade = true;
        previousLedMillis = currentMillis;
        break;

      case 16726215: // "OK" button, lights matrix
        digitalWrite(LED, HIGH);
        callMade = true;
        if (isLit) {
          color = strip.Color(0, 0, 0); // Lights off
          isLit = false;
        } else {
          isLit = true;
        }
        previousLedMillis = currentMillis;
        for (int i = 0; i < 64; i++) {
          strip.setPixelColor(i, color);
        }
        strip.show();
        color = strip.Color(100, 100, 100);
        break;

      // Feel free to add more cases for additional possible speed-dials, you'll have to purchase more lightning in your Blynk app, though!

      default:
        break;
    }
    Serial.println(value);
    irrecv.resume();
  }

  // Track the piezo sensor and change the matrix color accordingly. We call Blynk.run() several times here to make sure we stay connected to the Blynk server
  if (isLit) { // If the matrix is on
    if (currentMillis - previousPiezoMillis > 50) {
      piezoVal = analogRead(PIEZO);
      previousPiezoMillis = currentMillis;
      if (piezoVal > 30) { // The threshold of force needed to change color. Feel free to increase/decrease it if it suits you.
        Blynk.run();
        randNumber = random(0, 7);
        while (randNumber == prevNumber) { // Makes sure that the newly chosen color isn't the same as the current color
          randNumber = random(0, 7);
          Blynk.run();
        }
        prevNumber = randNumber;
        color = colors[randNumber]; // pick random color from colors array
        Blynk.run();
        for (int i = 0; i < 64; i++) { // display the chosen color on the matrix
          strip.setPixelColor(i, color);
          Blynk.run();
        }
        strip.show();
        Blynk.run();
      }
    }
  }
}

// This function gets us our GPS coordinates
BLYNK_WRITE(V0) {
  latitude = param[0].asFloat();
  longitude = param[1].asFloat();
  alt = param[2].asInt();
  s = param[3].asFloat();
}

void connect() {
  Serial.print("Connecting to Adafruit IO... ");
  int8_t ret;
  while ((ret = mqtt.connect()) != 0) {
    switch (ret) {
      case 1: Serial.println("Wrong protocol"); break;
      case 2: Serial.println("ID rejected"); break;
      case 3: Serial.println("Server unavail"); break;
      case 4: Serial.println("Bad user/pass"); break;
      case 5: Serial.println("Not authed"); break;
      case 6: Serial.println("Failed to subscribe"); break;
      default: Serial.println("Connection failed"); break;
    }
    if (ret >= 0) {
      mqtt.disconnect();
    }
    Serial.println("Retrying connection...");
    delay(5000);
  }
  Serial.println("Adafruit IO Connected!");
}
